import logging

import click
from geoseeq.knex import DEFAULT_ENDPOINT

from .. import Knex

logger = logging.getLogger('geoseeq_api')


class State(object):

    def __init__(self):
        self.api_token = None
        self.endpoint = DEFAULT_ENDPOINT
        self.outfile = None
        self.log_level = 20

    def get_knex(self):
        logger.setLevel(self.log_level)
        knex = Knex(self.endpoint)
        if self.api_token:
            knex.add_api_token(self.api_token)
        return knex


pass_state = click.make_pass_decorator(State, ensure=True)


def log_level_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        state.log_level = int(value)
        return value
    return click.option('-l', '--log-level',
                        type=int,
                        default=20,
                        envvar='GEOSEEQ_CLI_LOG_LEVEL',
                        expose_value=False,
                        callback=callback)(f)


def api_token_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        state.api_token = str(value)
        return value
    return click.option('-a', '--api-token',
                        envvar='GEOSEEQ_API_TOKEN',
                        expose_value=False,
                        help='Your GEOSEEQ API token.',
                        callback=callback)(f)


def endpoint_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        state.endpoint = str(value)
        return value
    return click.option('--endpoint',
                        default=DEFAULT_ENDPOINT,
                        expose_value=False,
                        help='The URL to use for GEOSEEQ.',
                        callback=callback)(f)


def outfile_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        state.outfile = value
        return value
    return click.option('-o', '--outfile',
                        default='-', type=click.File('w'),
                        expose_value=False,
                        help='Output file path',
                        callback=callback)(f)


def common_options(f):
    f = outfile_option(f)
    f = api_token_option(f)
    f = log_level_option(f)
    f = endpoint_option(f)
    return f


def use_common_state(f):
    f = common_options(f)
    f = pass_state(f)
    return f


def is_uuid(name):
    chunks = name.split('-')
    return len(chunks) == 5




from .blob_constructors import sample_from_uuid, sample_group_from_uuid
from .remote_object import RemoteObject


class WorkOrderProto(RemoteObject):
    remote_fields = [
        'uuid',
        'created_at',
        'updated_at',
        'name',
    ]
    parent_field = None

    def __init__(self, knex, uuid):
        super().__init__(self)
        self.knex = knex
        self.uuid = uuid
        self.cached_active_work_orders = []

    def get_active_work_orders(self, max_num=0, random=False, not_status=''):
        if self.cached_active_work_orders:
            return self.cached_active_work_orders
        url = f'work_order_prototypes/{self.uuid}/work_orders'
        url_options = {}
        if max_num > 0:
            url_options['max_num'] = max_num
        if random:
            url_options['random'] = 'random'
        if not_status:
            url_options['not_status'] = not_status
        blob = self.knex.get(url, url_options=url_options)
        for wo_blob in blob['results']:
            self.cached_active_work_orders.append(WorkOrder.from_blob(self.knex, wo_blob))
        return self.cached_active_work_orders

    def get_active_work_order_for_sample(self, sample):
        url = f'samples/{sample.uuid}/work_orders'
        blob = self.knex.get(url)
        for wo_blob in blob['results']:
            if wo_blob['name'] == self.name:  # TODO should be uuid but currently not a field
                wo = WorkOrder.from_blob(self.knex, wo_blob)
                wo._already_fetched = True
                wo._modified = False
                return wo
        raise KeyError(f'WorkOrder from Proto {self} not found for sample {sample}')

    def create_work_order_for_sample(self, sample):
        url = f'samples/{sample.uuid}/work_orders/{self.uuid}'
        response = self.knex.post(url)
        wo = WorkOrder.from_uuid(self.knex, response['uuid'])
        return wo

    def __str__(self):
        return f'<Geoseeq::WorkOrderProto {self.name} {self.uuid} />'

    def __repr__(self):
        return f'<Geoseeq::WorkOrderProto {self.name} {self.uuid} />'

    @classmethod
    def from_uuid(cls, knex, uuid):
        obj = cls(knex, uuid)
        blob = knex.get(f'work_order_prototypes/{uuid}')
        obj.load_blob(blob)
        return obj

    @classmethod
    def from_name(cls, knex, name):
        return cls.from_uuid(knex, name)


class WorkOrder(RemoteObject):
    remote_fields = [
        'uuid',
        'created_at',
        'updated_at',
        'name',
        'job_order_objs',
        'priority',
        'sample',
        'status',
    ]
    parent_field = None

    def __init__(self, knex, name):
        super().__init__(self)
        self.knex = knex
        self.name = name

    def get_sample(self):
        obj = sample_from_uuid(self.knex, self.sample)
        obj.url_options['work_order_uuid'] = self.uuid
        return obj

    def get_job_orders(self):
        for job_order_blob in self.job_order_objs:
            obj = JobOrder.from_blob(self, job_order_blob)
            obj._already_fetched = True
            obj._modified = False
            yield obj

    def get_job_order_by_name(self, name):
        jos = {jo.name: jo for jo in self.get_job_orders()}
        jo = jos[name]
        return jo

    def __str__(self):
        return f'<Geoseeq::WorkOrder {self.name} {self.uuid} />'

    def __repr__(self):
        return f'<Geoseeq::WorkOrder {self.name} {self.uuid} />'

    @classmethod
    def from_blob(cls, knex, blob):
        obj = cls(knex, blob['name'])
        obj.blob = blob
        obj.load_blob(blob)
        return obj

    @classmethod
    def from_uuid(cls, knex, uuid):
        url = f'work_orders/{uuid}'
        response = knex.get(url)
        return cls.from_blob(knex, response)


class JobOrder(RemoteObject):
    remote_fields = [
        'uuid',
        'created_at',
        'updated_at',
        'name',
        'analysis_result',
        'status',
    ]
    parent_field = None

    def __init__(self, knex, work_order):
        super().__init__(self)
        self.knex = knex
        self.work_order = work_order

    def _save(self):
        data = {
            field: getattr(self, field)
            for field in self.remote_fields if hasattr(self, field)
        }
        url = f'job_orders/{self.uuid}'
        self.knex.patch(url, json=data)

    def _get(self):
        blob = self.get_cached_blob()
        if not blob:
            blob = self.knex.get(f'job_orders/{self.uuid}')
            self.load_blob(blob)
            self.cache_blob(blob)
        else:
            self.load_blob(blob)

    def __str__(self):
        return f'<Geoseeq::JobOrder {self.name} {self.uuid} />'

    def __repr__(self):
        return f'<Geoseeq::JobOrder {self.name} {self.uuid} />'

    def pre_hash(self):
        key = self.work_order.uuid + self.name if self.name else ''
        return key

    @classmethod
    def from_blob(cls, work_order, blob):
        obj = cls(work_order.knex, work_order)
        obj.blob = blob
        obj.load_blob(blob)
        return obj

    @classmethod
    def from_uuid(cls, work_order, uuid):
        url = f'job_orders/{uuid}'
        response = work_order.knex.get(url)
        return cls.from_blob(work_order, response)


class GroupWorkOrderProto(RemoteObject):
    remote_fields = [
        'uuid',
        'created_at',
        'updated_at',
        'name',
    ]
    parent_field = None

    def __init__(self, knex, uuid):
        super().__init__(self)
        self.knex = knex
        self.uuid = uuid

    def get_active_work_orders(self):
        url = f'group_work_order_prototypes/{self.uuid}/work_orders'
        blob = self.knex.get(url)
        for wo_blob in blob['results']:
            yield GroupWorkOrder.from_blob(self.knex, wo_blob)

    def get_active_work_order_for_sample_group(self, group):
        for wo in self.get_active_work_orders():
            if wo.sample_group == group.uuid:
                return wo
        raise KeyError(f'WorkOrder from Proto {self} not found for group {group}')

    def create_work_order_for_sample_group(self, group):
        url = f'sample_groups/{group.uuid}/work_orders/{self.uuid}'
        response = self.knex.post(url)
        wo = GroupWorkOrder.from_uuid(self.knex, response['uuid'])
        return wo

    def __str__(self):
        return f'<Geoseeq::GroupWorkOrderProto {self.name} {self.uuid} />'

    def __repr__(self):
        return f'<Geoseeq::GropWorkOrderProto {self.name} {self.uuid} />'

    @classmethod
    def from_uuid(cls, knex, uuid):
        obj = cls(knex, uuid)
        blob = knex.get(f'group_work_order_prototypes/{uuid}')
        obj.load_blob(blob)
        return obj

    @classmethod
    def from_name(cls, knex, name):
        return cls.from_uuid(knex, name)


class GroupWorkOrder(RemoteObject):
    remote_fields = [
        'uuid',
        'created_at',
        'updated_at',
        'name',
        'work_order_links',
        'priority',
        'sample_group',
        'status',
    ]
    parent_field = None

    def __init__(self, knex, name):
        super().__init__(self)
        self.knex = knex
        self.name = name

    def get_sample_group(self):
        obj = sample_group_from_uuid(self.knex, self.sample_group)
        obj.url_options['work_order_uuid'] = self.uuid
        return obj

    def get_work_orders(self):
        for work_order_link in self.work_order_links:
            obj = WorkOrder.from_uuid(self, self.knex, work_order_link['uuid'])
            yield obj

    def __str__(self):
        return f'<Geoseeq::GroupWorkOrder {self.name} {self.uuid} />'

    def __repr__(self):
        return f'<Geoseeq::GroupWorkOrder {self.name} {self.uuid} />'

    @classmethod
    def from_blob(cls, knex, blob):
        obj = cls(knex, blob['name'])
        obj.blob = blob
        obj.load_blob(blob)
        return obj

    @classmethod
    def from_uuid(cls, knex, uuid):
        url = f'group_work_orders/{uuid}'
        response = knex.get(url)
        return cls.from_blob(knex, response)




import logging

import click

from .add import cli_add
from .copy import cli_copy
from .create import cli_create
from .delete import cli_delete
from .download import cli_download
from .list import cli_list
from .upload import cli_upload
from .user import cli_user
from .view import cli_view
from geoseeq.vc.cli import cli_vc

logger = logging.getLogger('geoseeq_api')
logger.addHandler(logging.StreamHandler())


@click.group()
def main():
    pass


@main.command()
def version():
    """Print the version of the Geoseeq API being used."""
    click.echo('0.1.5')  # remember to update setup



main.add_command(cli_add)
main.add_command(cli_create)
main.add_command(cli_download)
main.add_command(cli_list)
main.add_command(cli_upload)
main.add_command(cli_user)
main.add_command(cli_delete)
main.add_command(cli_copy)
main.add_command(cli_vc)
main.add_command(cli_view)



import json
import logging
import os
import time
from os.path import basename, getsize, join
from pathlib import Path
from tempfile import NamedTemporaryFile
from urllib.request import urlretrieve

import requests

from .constants import FIVE_MB
from .remote_object import RemoteObject, RemoteObjectError
from .utils import md5_checksum

logger = logging.getLogger("geoseeq_api")  # Same name as calling module
logger.addHandler(logging.NullHandler())  # No output unless configured by calling program


def diff_dicts(blob1, blob2):
    for problem in _diff_dicts("original", "$", blob1, blob2):
        yield problem
    for problem in _diff_dicts("serialized", "$", blob2, blob1):
        yield problem


def _diff_lists(suffix, depth, l1, l2):
    if len(l1) != len(l2):
        yield (f"MISMATCHED_LENGTHS:{suffix}:{depth}", len(l1), len(l2))
    for i in range(len(l1)):
        v1, v2 = l1[i], l2[i]
        if v1 != v2:
            mydepth = f"{depth}.index_{i}"
            if not isinstance(v1, type(v2)):
                yield (f"MISMATCHED_TYPES:{suffix}:{mydepth}", v1, v2)
            elif isinstance(v1, dict):
                for problem in _diff_dicts(suffix, mydepth, v1, v2):
                    yield problem
            elif isinstance(v1, list):
                for problem in _diff_lists(suffix, mydepth, v1, v2):
                    yield problem
            else:
                yield (f"MISMATCHED_LIST_VALUES:{suffix}:{mydepth}", v1, v2)


def _diff_dicts(suffix, depth, blob1, blob2):
    for k, v in blob1.items():
        if k not in blob2:
            yield (f"MISSING_KEY:{suffix}:{depth}", k, None)
        elif v != blob2[k]:
            if not isinstance(v, type(blob2[k])):
                yield (f"MISMATCHED_TYPES:{suffix}:{depth}", v, blob2[k])
            elif isinstance(v, dict):
                for problem in _diff_dicts(suffix, depth + "." + k, v, blob2[k]):
                    yield problem
            elif isinstance(v, list):
                for problem in _diff_lists(suffix, depth + "." + k, v, blob2[k]):
                    yield problem
            else:
                yield (f"MISMATCHED_DICT_VALUES:{suffix}:{depth}", v, blob2[k])


def check_json_serialization(blob):
    """Raise an error if serialization+deserialization fails to return an exact copy."""
    json_serialized = json.loads(json.dumps(blob))
    if json_serialized == blob:
        return
    if isinstance(blob, dict):
        issues = list(diff_dicts(blob, json_serialized))
    else:
        issues = [("MISMATCHED_VALUES:values_only:0", blob, json_serialized)]
    issues = [str(el) for el in issues]
    issues = "\n".join(issues)
    raise RemoteObjectError(f"JSON Serialization modifies object\nIssues:\n{issues}")


class AnalysisResult(RemoteObject):
    remote_fields = [
        "uuid",
        "created_at",
        "updated_at",
        "module_name",
        "replicate",
        "metadata",
        "description",
        "is_private",
        "pipeline_module",
    ]

    def _get(self):
        """Fetch the result from the server."""
        self.parent.idem()
        logger.debug(f"Getting AnalysisResult.")
        blob = self.get_cached_blob()
        if not blob:
            url = self.nested_url()
            if self.replicate:
                url += f"?replicate={self.replicate}"
            blob = self.knex.get(url, url_options=self.inherited_url_options)
            self.load_blob(blob)
            self.cache_blob(blob)
        else:
            self.load_blob(blob)

    def pre_hash(self):
        key = self.module_name + self.parent.pre_hash()
        key += self.replicate if self.replicate else ""
        return key

    def copy(self, new_parent, save=True):
        copied = new_parent.analysis_result(
            self.module_name, replicate=self.replicate, metadata=self.metadata
        )
        for field in self.get_fields():
            field.copy(copied, save=save)
        if save:
            copied.idem()
        return copied


class SampleAnalysisResult(AnalysisResult):
    parent_field = "sample"

    def __init__(self, knex, sample, module_name, replicate=None, metadata={}, is_private=False):
        super().__init__(self)
        self.knex = knex
        self.sample = sample
        self.parent = self.sample
        self.module_name = module_name
        self.replicate = replicate
        self._get_field_cache = []
        self.metadata = metadata
        self.is_private = is_private

    def nested_url(self):
        return self.sample.nested_url() + f"/analysis_results/{self.module_name}"

    def _save(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        data["sample"] = self.sample.uuid
        url = f"sample_ars/{self.uuid}"
        d = {"data": data, "url": url, "sample_ar": self}
        logger.debug(f"Saving SampleAnalysisResult. {d}")
        self.knex.put(url, json=data, url_options=self.inherited_url_options)
    
    def get_post_data(self):
        """Return a dict that can be used to POST this result to the server."""
        data = {
            field: getattr(self, field)
            for field in self.remote_fields
            if hasattr(self, field) and getattr(self, field) is not None
        }
        data["sample"] = self.sample.uuid
        if self.replicate:
            data["replicate"] = self.replicate
        return data

    def _create(self):
        self.sample.idem()
        data = self.get_post_data()
        d = {"data": data, "sample_ar": self}
        logger.debug(f"Creating SampleAnalysisResult. {d}")
        blob = self.knex.post(
            f"sample_ars?format=json", json=data, url_options=self.inherited_url_options
        )
        self.load_blob(blob)

    def field(self, field_name, data={}):
        d = {"data": data, "field_name": field_name, "sample_ar": self}
        logger.debug(f"Creating SampleAnalysisResultField for SampleAnalysisResult. {d}")
        return SampleAnalysisResultField(self.knex, self, field_name, data=data)

    def get_fields(self, cache=True):
        """Return a list of ar-fields fetched from the server."""
        if cache and self._get_field_cache:
            for field in self._get_field_cache:
                yield field
            return
        url = f"sample_ar_fields?analysis_result_id={self.uuid}"
        #url = self.nested_url() + f"/fields"
        logger.debug(f"Fetching SampleAnalysisResultFields. {self}")
        result = self.knex.get(url)
        for result_blob in result["results"]:
            result = self.field(result_blob["name"])
            result.load_blob(result_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            result._already_fetched = True
            result._modified = False
            if cache:
                self._get_field_cache.append(result)
            else:
                yield result
        if cache:
            for field in self._get_field_cache:
                yield field

    def __str__(self):
        return f"<Geoseeq::SampleResult {self.module_name} {self.replicate} {self.uuid} />"


class SampleGroupAnalysisResult(AnalysisResult):
    parent_field = "grp"

    def __init__(self, knex, grp, module_name, replicate=None, metadata={}, is_private=False):
        super().__init__(self)
        self.knex = knex
        self.grp = grp
        self.parent = self.grp
        self.module_name = module_name
        self.replicate = replicate
        self.metadata = metadata
        self.is_private = is_private

    def nested_url(self):
        return self.grp.nested_url() + f"/analysis_results/{self.module_name}"

    def _save(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        data["sample_group"] = self.grp.uuid
        url = f"sample_group_ars/{self.uuid}"
        self.knex.put(url, json=data)

    def _create(self):
        self.grp.idem()
        data = {
            "sample_group": self.grp.uuid,
            "module_name": self.module_name,
        }
        if self.replicate:
            data["replicate"] = self.replicate
        blob = self.knex.post(f"sample_group_ars?format=json", json=data)
        self.load_blob(blob)

    def field(self, field_name, data={}):
        return SampleGroupAnalysisResultField(self.knex, self, field_name, data=data)

    def get_fields(self):
        """Return a list of ar-fields fetched from the server."""
        url = f"sample_group_ar_fields?analysis_result_id={self.uuid}"
        result = self.knex.get(url)
        for result_blob in result["results"]:
            result = self.field(result_blob["name"])
            result.load_blob(result_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            result._already_fetched = True
            result._modified = False
            yield result

    def __str__(self):
        return f"<Geoseeq::SampleGroupResult {self.module_name} {self.replicate} {self.uuid} />"


class AnalysisResultField(RemoteObject):
    remote_fields = [
        "uuid",
        "created_at",
        "updated_at",
        "name",
        "stored_data",
    ]
    parent_field = "parent"

    def __init__(self, knex, parent, field_name, data={}):
        super().__init__(self)
        self.knex = knex
        self.parent = parent
        self.name = field_name
        self.stored_data = data
        self._cached_filename = None  # Used if the field points to S3, FTP, etc
        self._temp_filename = False

    @property
    def brn(self):
        obj_type = 'sample' if self.canon_url() == 'sample_ar_fields' else 'project'
        brn = f'brn:{self.knex.instance_code()}:{obj_type}_result_field:{self.uuid}'

    def nested_url(self):
        return self.parent.nested_url() + f"/fields/{self.name}"

    def get_blob_filename(self):
        sname = self.parent.parent.name.replace(".", "-")
        mname = self.parent.module_name.replace(".", "-")
        fname = self.name.replace(".", "-")
        filename = join(self.parent.parent.name, f"{sname}.{mname}.{fname}.json").replace(
            "::", "__"
        )
        return filename

    def get_referenced_filename_ext(self):
        try:
            key = [k for k in ["filename", "uri", "url"] if k in self.stored_data][0]
        except KeyError:
            raise TypeError("Cannot make a reference filename for a BLOB type result field.")
        ext = self.stored_data[key].split(".")[-1]
        if ext in ["gz"]:
            ext = self.stored_data[key].split(".")[-2] + "." + ext
        return ext

    def get_referenced_filename(self):
        ext = self.get_referenced_filename_ext()
        sname = self.parent.parent.name.replace(".", "-")
        mname = self.parent.module_name.replace(".", "-")
        fname = self.name.replace(".", "-")
        filename = join(self.parent.parent.name, f"{sname}.{mname}.{fname}.{ext}").replace(
            "::", "__"
        )
        return filename

    def get_local_filename(self):
        """Return a filename that can be used to store this field locally."""
        try:
            return basename(self.get_referenced_filename())
        except TypeError:
            return basename(self.get_blob_filename())

    def _save(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        data["analysis_result"] = self.parent.uuid
        url = f"{self.canon_url()}/{self.uuid}"
        self.knex.put(url, json=data)

    def _get(self):
        """Fetch the result from the server."""
        self.parent.idem()
        blob = self.knex.get(self.nested_url())
        self.load_blob(blob)

    def get_post_data(self):
        """Return a dict that can be used to POST this field to the server."""
        data = {
            "analysis_result": self.parent.uuid,
            "name": self.name,
            "stored_data": self.stored_data,
        }
        return data

    def link_file(self, link_type, *args, **kwargs):
        if link_type == "s3":
            return self.link_s3(*args, **kwargs)
        elif link_type == "ftp":
            return self.link_ftp(*args, **kwargs)
        elif link_type == "sra":
            return self.link_sra(*args, **kwargs)
        elif link_type == "azure":
            return self.link_azure(*args, **kwargs)

    def link_s3(self, url, endpoint_url=None):
        """Link this field to an S3 object.

        Args:
            url (str): The URL of the S3 object.
            endpoint_url (str): The URL of the S3 endpoint.
        """
        if endpoint_url is None:
            if not url.startswith("https://"):
                raise ValueError("endpoint_url must be specified for non-HTTPS URLs.")
            endpoint_url = "https://" + url.split("https://")[1].split("/")[0]
            url = "s3://" + url.split(endpoint_url)[1][1:]
        self.stored_data = {
            "__type__": "s3",
            "uri": url,
            "endpoint_url": endpoint_url,
        }
        return self.save()
    
    def link_ftp(self, url):
        """Link this field to an FTP object.

        Args:
            url (str): The URL of the FTP object.
        """
        self.stored_data = {
            "__type__": "ftp",
            "uri": url,
        }
        return self.save()

    def link_sra(self, url):
        """Link this field to an SRA object.

        Args:
            url (str): The URL of the SRA object.
        """
        self.stored_data = {
            "__type__": "sra",
            "uri": url,
        }
        return self.save()

    def link_azure(self, url):
        """Link this field to an Azure object.

        Args:
            url (str): The URL of the Azure object.
        """
        endpoint_url = "https://" + url.split("https://")[1].split("/")[0]
        self.stored_data = {
            "__type__": "azure",
            "uri": url,
            "endpoint_url": endpoint_url,
        }
        return self.save()

    def _create(self):
        check_json_serialization(self.stored_data)
        self.parent.idem()
        data = self.get_post_data()
        blob = self.knex.post(f"{self.canon_url()}?format=json", json=data)
        self.load_blob(blob)

    def get_download_url(self):
        """Return a URL that can be used to download the file for this result."""
        blob_type = self.stored_data.get("__type__", "").lower()
        if blob_type not in ["s3", "sra"]:
            raise TypeError("Cannot fetch a file for a BLOB type result field.")
        if blob_type == "s3":
            try:
                url = self.stored_data["presigned_url"]
            except KeyError:
                url = self.stored_data["uri"]
            if url.startswith("s3://"):
                url = self.stored_data["endpoint_url"] + "/" + url[5:]
            return url
        elif blob_type == "sra":
            url = self.stored_data["url"]
            return url

    def download_file(self, filename=None, cache=True):
        """Return a local filepath to the file this result points to."""
        blob_type = self.stored_data.get("__type__", "").lower()
        if blob_type not in ["s3", "sra", "ftp"]:
            raise TypeError("Cannot fetch a file for a BLOB type result field.")
        if cache and self._cached_filename:
            return self._cached_filename
        if blob_type == "s3":
            return self._download_s3(filename, cache)
        elif blob_type == "sra":
            return self._download_sra(filename, cache)
        elif blob_type == "ftp":
            return self._download_ftp(filename, cache)

    def _download_s3(self, filename, cache):
        try:
            url = self.stored_data["presigned_url"]
        except KeyError:
            url = self.stored_data["uri"]
        if url.startswith("s3://"):
            url = self.stored_data["endpoint_url"] + "/" + url[5:]
        if not filename:
            self._temp_filename = True
            myfile = NamedTemporaryFile(delete=False)
            myfile.close()
            filename = myfile.name
        urlretrieve(url, filename)
        if cache:
            self._cached_filename = filename
        return filename

    def _download_sra(self, filename, cache):
        return self._download_generic_url(filename, cache)

    def _download_ftp(self, filename, cache):
        return self._download_generic_url(filename, cache)

    def _download_generic_url(self, filename, cache):
        url = self.stored_data["url"]
        if not filename:
            self._temp_filename = True
            myfile = NamedTemporaryFile(delete=False)
            myfile.close()
            filename = myfile.name
        urlretrieve(url, filename)
        if cache:
            self._cached_filename = filename
        return filename

    # DEV: to simplify the uplaod process we will use only the multipart upload. It works well for small files also.
    # This function is currently unused.
    def upload_small_file(self, filepath, optional_fields={}):
        url = f"/{self.canon_url()}/{self.uuid}/upload_s3"
        filename = basename(filepath)
        optional_fields.update(
            {
                "md5_checksum": md5_checksum(filepath),
                "file_size_bytes": getsize(filepath),
            }
        )
        data = {
            "filename": filename,
            "optional_fields": optional_fields,
        }
        response = self.knex.post(url, json=data)
        with open(filepath, "rb") as f:
            files = {"file": (filename, f)}
            requests.post(  # Not a call to geoseeq so we do not use knex
                response["url"], data=response["fields"], files=files
            )
        return self

    def multipart_upload_file(
        self,
        filepath,
        file_size,
        optional_fields={},
        chunk_size=FIVE_MB,
        max_retries=3,
        logger=lambda x: x,
    ):
        n_parts = int(file_size / chunk_size) + 1
        optional_fields.update(
            {
                "md5_checksum": md5_checksum(filepath),
                "file_size_bytes": getsize(filepath),
            }
        )
        data = {
            "filename": basename(filepath),
            "optional_fields": optional_fields,
        }
        response = self.knex.post(f"/ar_fields/{self.uuid}/create_s3_upload", json=data)
        upload_id = response["upload_id"]
        parts = [*range(1, n_parts + 1)]
        data = {
            "parts": parts,
            "stance": 'upload-multipart',
            "upload_id": upload_id,
        }
        response = self.knex.post(f"/ar_fields/{self.uuid}/create_upload_urls", json=data)
        urls = response
        complete_parts = []
        logger(f'[INFO] Starting upload for "{filepath}"')
        with open(filepath, "rb") as f:
            for num, url in enumerate(list(urls.values())):
                file_data = f.read(chunk_size)
                attempts = 0
                while attempts < max_retries:
                    try:
                        http_response = requests.put(url, data=file_data)
                        http_response.raise_for_status()
                        break
                    except requests.exceptions.HTTPError:
                        logger(f"[WARN] Upload for part {num + 1} failed. Attempt {attempts + 1}")
                        attempts += 1
                        if attempts == max_retries:
                            raise
                        time.sleep(10**attempts)  # exponential backoff, (10 ** 2)s default max
                complete_parts.append({"ETag": http_response.headers["ETag"], "PartNumber": num + 1})
                logger(f'[INFO] Uploaded part {num + 1} of {len(urls)} for "{filepath}"')
        response = self.knex.post(
            f"/ar_fields/{self.uuid}/complete_upload_s3",
            json={
                "parts": complete_parts,
                "upload_id": upload_id,
            },
            json_response=False,
        )
        logger(f'[INFO] Finished Upload for "{filepath}"')
        return self

    def upload_file(self, filepath, multipart_thresh=FIVE_MB, **kwargs):
        resolved_path = Path(filepath).resolve()
        file_size = getsize(resolved_path)
        return self.multipart_upload_file(filepath, file_size, **kwargs)

    def __del__(self):
        if self._temp_filename and self._cached_filename:
            os.remove(self._cached_filename)

    def pre_hash(self):
        return self.name + self.parent.pre_hash()

    def copy(self, new_parent, save=True):
        copied = new_parent.field(self.name, data=self.stored_data)
        if save:
            copied.idem()
        return copied

    def checksum(self):
        """Return a checksum for this field as a blob.
        
        TODO
        """
        return {'value': '', 'method': 'none'}


class SampleAnalysisResultField(AnalysisResultField):
    
    def canon_url(self):
        return "sample_ar_fields"

    def __str__(self):
        return f"<Geoseeq::SampleResultField {self.name} {self.uuid} />"

class SampleGroupAnalysisResultField(AnalysisResultField):

    def canon_url(self):
        return "sample_group_ar_fields"

    def __str__(self):
        return f"<Geoseeq::SampleGroupResultField {self.name} {self.uuid} />"


import hashlib
import logging

from .file_system_cache import FileSystemCache

logger = logging.getLogger('geoseeq_api')  # Same name as calling module
logger.addHandler(logging.NullHandler())  # No output unless configured by calling program


def paginated_iterator(knex, initial_url, error_handler=None):
    cache = FileSystemCache()
    result = cache.get_cached_blob(initial_url)
    if not result:
        try:
            result = knex.get(initial_url)
        except Exception as e:
            logger.debug(f'Error fetching blob:\n\t{initial_url}\n\t{e}')
            if error_handler:
                error_handler(e)
            else:
                raise
        cache.cache_blob(initial_url, result)
    for blob in result['results']:
        yield blob
    next_page = result.get('next', None)
    if next_page:
        for blob in paginated_iterator(knex, next_page):
            yield blob


def md5_checksum(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()




import json
import logging

import click
from Bio import Entrez


from geoseeq.cli.utils import use_common_state
from geoseeq import Knex, Organization
from .api import create_geoseeq_group_from_bioproj
from .bioproject import BioProject, SRARecord
from .setup_logging import logger

logger.setLevel(10)
logger.handlers = []
logger.addHandler(logging.StreamHandler())


@click.group('ncbi')
def ncbi_main():
    pass


@ncbi_main.group('link')
def cli_ncbi_link():
    pass


@cli_ncbi_link.command('bioproject')
@use_common_state
@click.argument('org_name')
@click.argument('bioproj_accession')
def cli_ncbi_link_bioproject(state, bioproj_accession):
    """Create a geoseeq group from an NCBI BioProject.

    Creates a GeoSeeq SampleGroup corresponding to the given bioproject accession.
     - A GeoSeeq Sample is created in the group for each NCBI BioSample
     - raw read analysis results are created for each sample from the SRA

    All objects are tagged to flag they are from an NCBI database.
    """
    logger.info(f'Creating GeoSeeq SampleGroup from BioProject "{bioproj_accession}"')
    Entrez.email = email
    knex = Knex(endpoint)
    if api_token:
        knex.add_api_token(api_token)
    org = Organization(knex, org_name).get()
    grp = create_geoseeq_group_from_bioproj(org, bioproj_accession)
    return grp


@ncbi_main.group('list')
def cli_ncbi_list():
    pass


@cli_ncbi_list.command('bioproject')
@click.option('-s', '--sleep', default=1)
@click.option('-o', '--outfile', type=click.File('w'), default='-')
@click.argument('accession')
def cli_ncbi_list_bioproject(sleep, outfile, accession):
    """Write sample info from an NCBI bioproject to a file."""

    bioproj = BioProject(accession)
    print(bioproj)
    print(
        json.dumps(bioproj.metadata(), indent=4, sort_keys=True),
        file=outfile
    )
    for biosample in bioproj.biosamples():
        print('\t', biosample)
        for sra_rec in biosample.sra():
            print('\t\t', sra_rec)
            assert False


@cli_ncbi_list.command('sra')
@click.option('-s', '--sleep', default=1)
@click.option('-o', '--outfile', type=click.File('w'), default='-')
@click.argument('accession')
def cli_ncbi_list_sra(sleep, outfile, accession):
    """Write sample info from an NCBI bioproject to a file."""
    sra = SRARecord(accession)
    tbl = sra.to_table(sleep=sleep)
    tbl.to_csv(outfile)



from .analysis_result import (
    SampleAnalysisResult,
    SampleGroupAnalysisResult,
    SampleAnalysisResultField,
    SampleGroupAnalysisResultField,
)
from .organization import Organization
from .sample import Sample
from .sample_group import SampleGroup
from .knex import GeoseeqNotFoundError


def org_from_blob(knex, blob, already_fetched=True, modified=False):
    org = Organization(knex, blob["name"])
    org.load_blob(blob)
    org._already_fetched = already_fetched
    org._modified = modified
    return org


def sample_group_from_blob(knex, blob, already_fetched=True, modified=False):
    org = org_from_blob(
        knex, blob["organization_obj"], already_fetched=already_fetched, modified=modified
    )
    grp = SampleGroup(knex, org, blob["name"], is_library=blob["is_library"])
    grp.load_blob(blob)
    grp._already_fetched = already_fetched
    grp._modified = modified
    return grp


def sample_from_blob(knex, blob, already_fetched=True, modified=False):
    lib = sample_group_from_blob(
        knex, blob["library_obj"], already_fetched=already_fetched, modified=modified
    )
    sample = Sample(knex, lib, blob["name"], metadata=blob["metadata"])
    sample.load_blob(blob)
    sample._already_fetched = already_fetched
    sample._modified = modified
    return sample


def sample_group_ar_from_blob(knex, blob, already_fetched=True, modified=False):
    group = sample_group_from_blob(
        knex, blob["sample_group_obj"], already_fetched=already_fetched, modified=modified
    )
    ar = SampleGroupAnalysisResult(
        knex, sample, blob["module_name"], replicate=blob["replicate"], metadata=blob["metadata"]
    )
    ar.load_blob(blob)
    ar._already_fetched = already_fetched
    ar._modified = modified
    return ar


def sample_ar_from_blob(knex, blob, already_fetched=True, modified=False):
    sample = sample_from_blob(
        knex, blob["sample_obj"], already_fetched=already_fetched, modified=modified
    )
    ar = SampleAnalysisResult(
        knex, sample, blob["module_name"], replicate=blob["replicate"], metadata=blob["metadata"]
    )
    ar.load_blob(blob)
    ar._already_fetched = already_fetched
    ar._modified = modified
    return ar


def sample_group_ar_from_blob(knex, blob, already_fetched=True, modified=False):
    group = sample_group_from_blob(
        knex, blob["sample_group_obj"], already_fetched=already_fetched, modified=modified
    )
    ar = SampleGroupAnalysisResult(
        knex, sample, blob["module_name"], replicate=blob["replicate"], metadata=blob["metadata"]
    )
    ar.load_blob(blob)
    ar._already_fetched = already_fetched
    ar._modified = modified
    return ar


def sample_ar_field_from_blob(knex, blob, already_fetched=True, modified=False):
    ar = sample_ar_from_blob(
        knex, blob["analysis_result_obj"], already_fetched=already_fetched, modified=modified
    )
    arf = SampleAnalysisResultField(
        knex, ar, blob["name"], data=blob["stored_data"]
    )
    arf.load_blob(blob)
    ar._already_fetched = already_fetched
    ar._modified = modified
    return arf


def sample_group_ar_field_from_blob(knex, blob, already_fetched=True, modified=False):
    ar = sample_group_ar_from_blob(
        knex, blob["analysis_result_obj"], already_fetched=already_fetched, modified=modified
    )
    arf = SampleGroupAnalysisResultField(
        knex, ar, blob["name"], data=blob["stored_data"]
    )
    arf.load_blob(blob)
    ar._already_fetched = already_fetched
    ar._modified = modified
    return arf


def sample_group_from_uuid(knex, uuid):
    blob = knex.get(f"sample_groups/{uuid}")
    sample = sample_group_from_blob(knex, blob)
    return sample


def sample_from_uuid(knex, uuid):
    blob = knex.get(f"samples/{uuid}")
    sample = sample_from_blob(knex, blob)
    return sample


def sample_ar_from_uuid(knex, uuid):
    blob = knex.get(f"sample_ars/{uuid}")
    ar = sample_ar_from_blob(knex, blob)
    return ar


def sample_group_ar_from_uuid(knex, uuid):
    blob = knex.get(f"sample_group_ars/{uuid}")
    ar = sample_group_ar_from_blob(knex, blob)
    return ar


def sample_ar_field_from_uuid(knex, uuid):
    blob = knex.get(f"sample_ar_fields/{uuid}")
    ar = sample_ar_field_from_blob(knex, blob)
    return ar


def sample_group_ar_field_from_uuid(knex, uuid):
    blob = knex.get(f"sample_group_ar_fields/{uuid}")
    ar = sample_group_ar_field_from_blob(knex, blob)
    return ar


def resolve_brn(knex, brn):
    """Return the object which the brn points to."""
    assert brn.startswith(f'brn:{knex.instance_code()}:')
    _, _, object_type, uuid = brn.split(':')
    if object_type == 'project':
        return object_type, sample_group_from_uuid(knex, uuid)
    if object_type == 'sample':
        return object_type, sample_from_uuid(knex, uuid)
    if object_type == 'sample_result':
        return object_type, sample_ar_from_uuid(knex, uuid)
    if object_type == 'sample_result_field':
        return object_type, sample_ar_field_from_uuid(knex, uuid)
    if object_type == 'project_result':
        return object_type, sample_group_ar_from_uuid(knex, uuid)
    if object_type == 'project_result_field':
        return object_type, sample_group_ar_field_from_uuid(knex, uuid)
    raise GeoseeqNotFoundError(f'Type "{object_type}" not found')



import logging

import click

from .. import Organization
from .utils import use_common_state

logger = logging.getLogger('geoseeq_api')


@click.group('delete')
def cli_delete():
    """Delete objects from GeoSeeq."""
    pass


@cli_delete.command('samples')
@use_common_state
@click.option('-c/-n', '--confirm/--no-confirm', default=True)
@click.argument('org_name')
@click.argument('grp_name')
@click.argument('sample_names', nargs=-1)
def cli_delete_samples(state, confirm, org_name, grp_name, sample_names):
    """Delete a list of samples in the specified group."""
    knex = state.get_knex()
    org = Organization(knex, org_name).get()
    grp = org.sample_group(grp_name).get()
    for sample_name in sample_names:
        sample = grp.sample(sample_name).get()
        if confirm and click.confirm(f'Delete Sample {sample}?'):
            sample.delete()
            logger.info(f'Deleted Sample {sample}')




from .blob_constructors import (
    sample_from_blob,
    sample_result_from_blob,
    sample_ar_field_from_blob,
)


def bulk_create_samples(knex, samples):
    """Create multiple samples at once. Returns a list of created samples.
    
    Only returns samples which were newly created.
    If a sample already exists on the server, it will not be returned.
    """
    result = knex.post(
        "bulk_samples",
        json={"samples": [sample.get_post_data() for sample in samples]},
    )
    created_samples = [
        sample_from_blob(knex, result_blob) for result_blob in result if result_blob
    ]
    return created_samples


def bulk_create_sample_results(knex, sample_results):
    """Create multiple sample results at once. Returns a list of created sample results.
    
    Only returns sample results which were newly created.
    If a sample result already exists on the server, it will not be returned.
    """
    result = knex.post(
        "bulk_sample_results",
        json={"sample_results": [sample_result.get_post_data() for sample_result in sample_results]},
    )
    created_sample_results = [
        sample_result_from_blob(knex, result_blob) for result_blob in result if result_blob
    ]
    return created_sample_results


def bulk_create_sample_result_fields(knex, sample_result_fields):
    """Create multiple sample result fields at once. Returns a list of created sample result fields.
    
    Only returns sample result fields which were newly created.
    If a sample result field already exists on the server, it will not be returned.    
    """
    result = knex.post(
        "bulk_sample_result_fields",
        json={"sample_result_fields": [sample_result_field.get_post_data() for sample_result_field in sample_result_fields]},
    )
    created_sample_result_fields = [
        sample_ar_field_from_blob(knex, result_blob) for result_blob in result if result_blob
    ]
    return created_sample_result_fields


import logging
import requests
from os import environ
from .file_system_cache import FileSystemCache

DEFAULT_ENDPOINT = "https://backend.geoseeq.com"


logger = logging.getLogger("geoseeq_api")  # Same name as calling module
logger.addHandler(logging.NullHandler())  # No output unless configured by calling program


def clean_url(url):
    if url[-1] == "/":
        url = url[:-1]
    return url


class TokenAuth(requests.auth.AuthBase):
    """Attaches MetaGenScope Token Authentication to the given Request object."""

    def __init__(self, token):
        self.token = token

    def __call__(self, request):
        """Add authentication header to request."""
        request.headers["Authorization"] = f"Token {self.token}"
        return request

    def __str__(self):
        """Return string representation of TokenAuth."""
        return self.token


class GeoseeqGeneralError(requests.exceptions.HTTPError):
    pass


class GeoseeqNotFoundError(GeoseeqGeneralError):
    pass


class GeoseeqForbiddenError(GeoseeqGeneralError):
    pass


class GeoseeqInternalError(GeoseeqGeneralError):
    pass

class GeoseeqTimeoutError(GeoseeqGeneralError):
    pass

class GeoseeqOtherError(GeoseeqGeneralError):
    pass


class Knex:

    def __init__(self, endpoint_url=DEFAULT_ENDPOINT):
        self.endpoint_url = endpoint_url
        self.endpoint_url += "/api"
        self.auth = None
        self.headers = {"Accept": "application/json"}
        self.cache = FileSystemCache()
        self._verify = self._set_verify()
        self.sess = self._new_session()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self.sess.close()

    def _new_session(self):
        if hasattr(self, 'sess') and self.sess:
            self.sess.close()
        sess = requests.Session()
        sess.headers = self.headers
        sess.auth = self.auth
        sess.verify = self._verify
        return sess

    def _set_verify(self):
        try:
            val = environ['GEOSEEQ_NO_SSL_VERIFICATION']
            if val.lower() == 'true':
                return False
            if val.lower() == 'false':
                return True
            return val
        except KeyError:
            return True


    def _logging_info(self, **kwargs):
        base = {"endpoint_url": self.endpoint_url, "headers": self.headers}
        base.update(kwargs)
        return base

    def _clean_url(self, url, url_options={}):
        url = clean_url(url)
        url = url.replace(self.endpoint_url, "")
        if url[0] == "/":
            url = url[1:]
        if url_options:
            opts = [f"{key}={val}" for key, val in url_options.items()]
            opts = "&".join(opts)
            if "?" in url:
                url += "&" + opts
            else:
                url += "?" + opts
        return url

    def add_api_token(self, token):
        self.auth = TokenAuth(token)
        self.sess = self._new_session()

    def _handle_response(self, response, json_response=True):
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            if response.status_code == 403:
                raise GeoseeqForbiddenError(e)
            if response.status_code == 404:
                raise GeoseeqNotFoundError(e)
            if response.status_code == 500:
                raise GeoseeqInternalError(e)
            if response.status_code == 504:
                raise GeoseeqTimeoutError(e)
            raise GeoseeqOtherError(e)
        except Exception:
            logger.debug(f"Request failed. {response}\n{response.content}")
            raise
        if json_response:
            return response.json()
        return response

    def instance_code(self):
        return "gsr1"  # TODO

    def get(self, url, url_options={}, **kwargs):
        url = self._clean_url(url, url_options=url_options)
        d = self._logging_info(url=url, auth_token=self.auth)
        logger.debug(f"Sending GET request. {d}")
        response = self.sess.get(f"{self.endpoint_url}/{url}")
        resp = self._handle_response(response, **kwargs)
        return resp

    def post(self, url, json={}, url_options={}, **kwargs):
        url = self._clean_url(url, url_options=url_options)
        d = self._logging_info(url=url, auth_token=self.auth, json=json)
        logger.debug(f"Sending POST request. {d}")
        response = self.sess.post(
            f"{self.endpoint_url}/{url}",
            json=json
        )
        return self._handle_response(response, **kwargs)

    def put(self, url, json={}, url_options={}, **kwargs):
        url = self._clean_url(url, url_options=url_options)
        d = self._logging_info(url=url, auth_token=self.auth, json=json)
        logger.debug(f"Sending PUT request. {d}")
        response = self.sess.put(
            f"{self.endpoint_url}/{url}",
            json=json,
        )
        return self._handle_response(response, **kwargs)

    def patch(self, url, json={}, url_options={}, **kwargs):
        url = self._clean_url(url, url_options=url_options)
        d = self._logging_info(url=url, auth_token=self.auth, json=json)
        logger.debug(f"Sending PATCH request. {d}")
        response = self.sess.patch(
            f"{self.endpoint_url}/{url}",
            json=json,
        )
        return self._handle_response(response, **kwargs)

    def delete(self, url, url_options={}, **kwargs):
        url = self._clean_url(url, url_options=url_options)
        d = self._logging_info(url=url, auth_token=self.auth)
        logger.debug(f"Sending DELETE request. {d}")
        response = self.sess.delete(f"{self.endpoint_url}/{url}")
        logger.debug(f"DELETE request response:\n{response}")
        return self._handle_response(response, json_response=False, **kwargs)




from .remote_object import RemoteObject


class Pipeline(RemoteObject):
    remote_fields = [
        'uuid',
        'created_at',
        'updated_at',
        'name',
        'description',
        'long_description',
    ]
    parent_field = None

    def __init__(self, knex, name):
        super().__init__(self)
        self.knex = knex
        self.name = name
        self.description = ''
        self.long_description = ''

    def _save(self):
        data = {
            field: getattr(self, field)
            for field in self.remote_fields if hasattr(self, field)
        }
        url = f'pipelines/{self.uuid}'
        self.knex.put(url, json=data)

    def _get(self):
        """Fetch the result from the server."""
        blob = self.get_cached_blob()
        if not blob:
            blob = self.knex.get(f'pipelines/name/{self.name}')
            self.load_blob(blob)
            self.cache_blob(blob)
        else:
            self.load_blob(blob)

    def _create(self):
        data = {
            'name': self.name,
            'description': self.description,
            'long_description': self.long_description,
        }
        url = 'pipelines?format=json'
        blob = self.knex.post(url, json=data)
        self.load_blob(blob)

    def __str__(self):
        return f'<Geoseeq::Pipeline {self.name} {self.uuid} />'

    def __repr__(self):
        return f'<Geoseeq::Pipeline {self.name} {self.uuid} />'

    def pre_hash(self):
        return 'PIPELINE' + self.name

    def module(self, name, version, metadata={}):
        return PipelineModule(
            self.knex,
            self,
            name,
            version,
            metadata=metadata,
        )

    def get_modules(self):
        url = f'pipeline_modules?pipeline={self.uuid}'
        result = self.knex.get(url)
        for result_blob in result['results']:
            result = self.module(result_blob['name'], result_blob['version'])
            result.load_blob(result_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            result._already_fetched = True
            result._modified = False
            yield result


class PipelineModule(RemoteObject):
    remote_fields = [
        'uuid',
        'created_at',
        'updated_at',
        'name',
        'version',
        'metadata',
        'description',
        'long_description',
        'dependencies',
    ]
    parent_field = 'pip'

    def __init__(self, knex, pipeline, name, version, metadata={}):
        super().__init__(self)
        self.knex = knex
        self.pip = pipeline
        self.name = name
        self.version = version
        self.metadata = metadata
        self.description = ''
        self.long_description = ''
        self.dependencies = []

    def _save(self):
        data = {
            field: getattr(self, field)
            for field in self.remote_fields if hasattr(self, field)
        }
        url = f'pipeline_modules/{self.uuid}'
        self.knex.put(url, json=data)

    def _get(self):
        """Fetch the result from the server."""
        blob = self.get_cached_blob()
        if not blob:
            blob = self.knex.get(f'pipelines/{self.pip.uuid}/modules/{self.name}/{self.version}')
            self.load_blob(blob)
            self.cache_blob(blob)
        else:
            self.load_blob(blob)

    def _create(self):
        data = {
            'pipeline': self.pip.uuid,
            'name': self.name,
            'version': self.version,
            'metadata': self.metadata,
            'description': self.description,
            'long_description': self.long_description,
        }
        url = 'pipeline_modules?format=json'
        blob = self.knex.post(url, json=data)
        self.load_blob(blob)

    def add_dependency(self, upstream):
        self.dependencies.append(upstream.uuid)

    def __str__(self):
        return f'<Geoseeq::PipelineModule "{self.name}" "{self.version}" {self.uuid} />'

    def __repr__(self):
        return f'<Geoseeq::PipelineModule "{self.name}" "{self.version}" {self.uuid} />'

    def pre_hash(self):
        return 'PIPELINE_MODULE' + self.name + self.version + self.pip.pre_hash()



import click

from .. import Organization
from .utils import use_common_state


@click.group('list')
def cli_list():
    """List objects on GeoSeeq."""
    pass


@cli_list.command('samples')
@use_common_state
@click.argument('org_name')
@click.argument('grp_name')
def cli_list_samples(state, org_name, grp_name):
    """Print a list of samples in the specified group."""
    knex = state.get_knex()
    org = Organization(knex, org_name).get()
    grp = org.sample_group(grp_name).get()
    for sample in grp.get_samples():
        print(sample, file=state.outfile)



"""Test suite for experimental functions."""
import random
from os import environ
from os.path import dirname, join
from unittest import TestCase, skip

from geoseeq_api import Knex, Organization, RemoteObjectError, SampleGroup
from requests.exceptions import HTTPError

PACKET_DIR = join(dirname(__file__), "built_packet")
ENDPOINT = environ.get("GEOSEEQ_API_TESTING_ENDPOINT", "http://127.0.0.1:8000")


def random_str(len=12):
    """Return a random alphanumeric string of length `len`."""
    out = random.choices("abcdefghijklmnopqrtuvwxyzABCDEFGHIJKLMNOPQRTUVWXYZ0123456789", k=len)
    return "".join(out)


class TestGeoseeqApiClient(TestCase):
    """Test suite for packet building."""

    def setUp(self):
        self.knex = Knex(ENDPOINT)
        # Creates a test user and an API token for the user in database. Returns the token.
        api_token = self.knex.post("/users/test-user", json={"email": f"{random_str()}@gmail.com"})
        self.knex.add_api_token(api_token)

    def test_create_org(self):
        """Test that we can create an org."""
        org = Organization(self.knex, f"my_client_test_org {random_str()}")
        org.create()
        self.assertTrue(org.uuid)

    def test_get_org(self):
        """Test that we can get an org."""
        name = f"my_client_test_org {random_str()}"
        Organization(self.knex, name).create()
        org = Organization(self.knex, name).get()
        self.assertTrue(org.uuid)

    def test_idem_create_org(self):
        """Test that we can create an org using idem."""
        org = Organization(self.knex, f"my_client_test_org {random_str()}")
        org.idem()
        self.assertTrue(org.uuid)

    def test_idem_get_org(self):
        """Test that we can get an org using idem."""
        name = f"my_client_test_org {random_str()}"
        Organization(self.knex, name).create()
        org = Organization(self.knex, name).idem()
        self.assertTrue(org.uuid)

    def test_create_group(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        # N.B. It should NOT be necessary to call org.create()
        grp = org.sample_group(f"my_client_test_grp {key}")
        grp.create()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)

    def test_create_library(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}").create()
        grp_name = f"my_client_test_grp {key}"
        org.sample_group(grp_name, is_library=True).create()
        grp = SampleGroup(self.knex, org, grp_name).get()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)
        self.assertTrue(grp.is_library)

    def test_create_group_result(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}")
        # N.B. It should NOT be necessary to call <parent>.create()
        ar = grp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        ar.create()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)
        self.assertTrue(ar.uuid)

    def test_create_group_result_field(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}")
        ar = grp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        # N.B. It should NOT be necessary to call <parent>.create()
        field = ar.field("my_client_test_field_name", {"foo": "bar"})
        field.create()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)
        self.assertTrue(ar.uuid)
        self.assertTrue(field.uuid)

    def test_create_sample(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        # N.B. It should NOT be necessary to call <parent>.create()
        samp = grp.sample(f"my_client_test_sample {key}")
        samp.create()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)
        self.assertTrue(samp.uuid)

    def test_add_sample(self):
        """Test that we can add a sample to a (non-library) sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        lib = org.sample_group(f"my_client_test_lib {key}", is_library=True)
        samp = lib.sample(f"my_client_test_sample {key}").create()

        grp = org.sample_group(f"my_client_test_grp {key}", is_library=False).create()
        grp.add_sample(samp).save()
        self.assertIn(samp.uuid, {samp.uuid for samp in grp.get_samples()})

    def test_get_samples_in_group(self):
        """Test that we can get the samples in a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp_names = [f"my_client_test_sample_{i} {key}" for i in range(10)]
        for samp_name in samp_names:
            grp.sample(samp_name).create()
        retrieved = org.sample_group(f"my_client_test_grp {key}", is_library=True).get()
        retrieved_names = set()
        for samp in retrieved.get_samples():
            retrieved_names.add(samp.name)
            self.assertTrue(samp.uuid)
        for samp_name in samp_names:
            self.assertIn(samp_name, retrieved_names)

    def test_get_results_in_group(self):
        """Test that we can get the results in a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        result_names = [("my_client_test_module", f"replicate_{i}") for i in range(10)]
        for module_name, replicate in result_names:
            grp.analysis_result(module_name, replicate=replicate).create()
        retrieved = org.sample_group(f"my_client_test_grp {key}", is_library=True).get()
        retrieved_names = set()
        for result in retrieved.get_analysis_results():
            retrieved_names.add((result.module_name, result.replicate))
            self.assertTrue(result.uuid)
        for result_name_rep in result_names:
            self.assertIn(result_name_rep, retrieved_names)

    def test_get_results_in_sample(self):
        """Test that we can get the results in a sample."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}").create()
        result_names = [("my_client_test_module", f"replicate_{i}") for i in range(10)]
        for module_name, replicate in result_names:
            samp.analysis_result(module_name, replicate=replicate).create()
        retrieved = grp.sample(f"my_client_test_sample {key}").get()
        retrieved_names = set()
        for result in retrieved.get_analysis_results():
            retrieved_names.add((result.module_name, result.replicate))
            self.assertTrue(result.uuid)
        for result_name_rep in result_names:
            self.assertIn(result_name_rep, retrieved_names)

    def test_get_result_fields(self):
        """Test that we can get the fields of an analysis result."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}")
        ar = samp.analysis_result("my_client_test_module").create()
        self.assertTrue(grp.uuid)

        field_names = [f"field_{i}" for i in range(10)]
        for field_name in field_names:
            ar.field(field_name).create()

        retrieved = samp.analysis_result("my_client_test_module").get()
        retrieved_names = set()
        for result in retrieved.get_fields():
            retrieved_names.add(result.name)
            self.assertTrue(result.uuid)
        for result_name_rep in field_names:
            self.assertIn(result_name_rep, retrieved_names)

    @skip("nonfunctional currently")
    def test_delete_sample(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        # N.B. It should NOT be necessary to call <parent>.create()
        samp = grp.sample(f"my_client_test_sample {key}")
        samp.create()
        self.assertTrue(samp.uuid)
        samp.delete()
        self.assertRaises(lambda: setattr(samp, "name", "foo"), RemoteObjectError)
        retrieved = grp.sample(f"my_client_test_sample {key}")
        self.assertRaises(retrieved.get, HTTPError)

    def test_modify_sample_save(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        # N.B. It should NOT be necessary to call <parent>.create()
        samp = grp.sample(f"my_client_test_sample {key}")
        samp.create()
        self.assertTrue(grp.is_public)
        self.assertTrue(samp.uuid)
        self.assertTrue(samp._already_fetched)
        self.assertFalse(samp._modified)
        samp.metadata = {f"metadata_{key}": "some_new_metadata"}
        self.assertTrue(samp._modified)
        samp.save()
        self.assertTrue(samp._already_fetched)
        self.assertFalse(samp._modified)
        retrieved = grp.sample(f"my_client_test_sample {key}").get()
        self.assertIn(f"metadata_{key}", retrieved.metadata)

    def test_modify_sample_idem(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        # N.B. It should NOT be necessary to call <parent>.create()
        samp = grp.sample(f"my_client_test_sample {key}")
        samp.create()
        self.assertTrue(samp.uuid)
        self.assertTrue(samp._already_fetched)
        self.assertFalse(samp._modified)
        samp.metadata = {f"metadata_{key}": "some_new_metadata"}
        self.assertTrue(samp._modified)
        samp.idem()
        self.assertTrue(samp._already_fetched)
        self.assertFalse(samp._modified)
        retrieved = grp.sample(f"my_client_test_sample {key}").get()
        self.assertIn(f"metadata_{key}", retrieved.metadata)

    def test_create_sample_result(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}")
        # N.B. It should NOT be necessary to call <parent>.create()
        ar = samp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        ar.create()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)
        self.assertTrue(samp.uuid)
        self.assertTrue(ar.uuid)

    def test_create_sample_result_field(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}")
        ar = samp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        # N.B. It should NOT be necessary to call <parent>.create()
        field = ar.field("my_client_test_field_name", {"foo": "bar"})
        field.create()
        self.assertTrue(org.uuid)
        self.assertTrue(grp.uuid)
        self.assertTrue(samp.uuid)
        self.assertTrue(ar.uuid)
        self.assertTrue(field.uuid)

    def test_modify_sample_result_field_idem(self):
        """Test that we can create a sample group."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}")
        ar = samp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        # N.B. It should NOT be necessary to call <parent>.create()
        field = ar.field(f"my_client_test_field_name {key}", {"foo": "bar"})
        field.create()
        self.assertTrue(field.uuid)
        field.stored_data = {"foo": "bizz"}  # TODO: handle deep modifications
        field.idem()
        retrieved = ar.field(f"my_client_test_field_name {key}").get()
        self.assertEqual(retrieved.stored_data["foo"], "bizz")

    def test_get_sample_group_manifest(self):
        """Test that we can get a group manifest."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}")
        ar = samp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        field = ar.field("my_client_test_field_name", {"foo": "bar"})
        field.create()
        manifest = grp.get_manifest()
        self.assertTrue(manifest)

    def test_get_sample_manifest(self):
        """Test that we can get a group manifest."""
        key = random_str()
        org = Organization(self.knex, f"my_client_test_org {key}")
        grp = org.sample_group(f"my_client_test_grp {key}", is_library=True)
        samp = grp.sample(f"my_client_test_sample {key}")
        ar = samp.analysis_result(f"my_client_test_module_name")  # no {key} necessary
        field = ar.field("my_client_test_field_name", {"foo": "bar"})
        field.create()
        manifest = samp.get_manifest()
        self.assertTrue(manifest)



import json
from os import makedirs
from os.path import dirname, join

import click
import pandas as pd

from .. import Organization
from .utils import use_common_state


@click.group("download")
def cli_download():
    """Download objects from GeoSeeq."""
    pass


def _setup_download(state, sample_manifest, org_name, grp_name, sample_names):
    knex = state.get_knex()
    org = Organization(knex, org_name).get()
    grp = org.sample_group(grp_name).get()
    if sample_manifest:
        sample_names = set(sample_names) | set([el.strip() for el in sample_manifest if el])
    return grp, sample_names


@cli_download.command("metadata")
@use_common_state
@click.option(
    "--sample-manifest", type=click.File("r"), help="List of sample names to download from"
)
@click.argument("org_name")
@click.argument("grp_name")
@click.argument("sample_names", nargs=-1)
def cli_download_metadata(state, sample_manifest, org_name, grp_name, sample_names):
    """Download Sample Analysis Results for a set of samples."""
    grp, sample_names = _setup_download(state, sample_manifest, org_name, grp_name, sample_names)
    metadata = {}
    for sample in grp.get_samples(cache=False):
        if sample_names and sample.name not in sample_names:
            continue
        metadata[sample.name] = sample.metadata
    metadata = pd.DataFrame.from_dict(metadata, orient="index")
    metadata.to_csv(state.outfile)
    click.echo("Metadata successfully downloaded for samples.", err=True)


@cli_download.command("sample-results")
@use_common_state
@click.option("--module-name")
@click.option("--field-name")
@click.option("--target-dir", default=".")
@click.option(
    "--sample-manifest",
    default=None,
    type=click.File("r"),
    help="List of sample names to download from",
)
@click.option("--download/--urls-only", default=True, help="Download files or just print urls")
@click.argument("org_name")
@click.argument("grp_name")
@click.argument("sample_names", nargs=-1)
def cli_download_sample_results(
    state,
    module_name,
    field_name,
    target_dir,
    sample_manifest,
    download,
    org_name,
    grp_name,
    sample_names,
):
    """Download Sample Analysis Results for a set of samples."""
    grp, sample_names = _setup_download(state, sample_manifest, org_name, grp_name, sample_names)
    for sample in grp.get_samples(cache=False):
        if sample_names and sample.name not in sample_names:
            continue
        for ar in sample.get_analysis_results(cache=False):
            if module_name and ar.module_name != module_name:
                continue
            for field in ar.get_fields(cache=False):
                if field_name and field.name != field_name:
                    continue
                if not download:  # download urls to a file, not actual files.
                    try:
                        print(
                            field.get_download_url(),
                            field.get_referenced_filename(),
                            file=state.outfile,
                        )
                    except TypeError:
                        pass
                    continue
                filename = join(target_dir, field.get_blob_filename()).replace("::", "__")
                makedirs(dirname(filename), exist_ok=True)
                click.echo(f"Downloading BLOB {sample} :: {ar} :: {field} to {filename}", err=True)
                with open(filename, "w") as blob_file:
                    blob_file.write(json.dumps(field.stored_data))
                try:
                    filename = join(target_dir, field.get_referenced_filename()).replace("::", "__")
                    click.echo(
                        f"Downloading FILE {sample} :: {ar} :: {field} to {filename}", err=True
                    )
                    field.download_file(filename=filename)
                except TypeError:
                    pass
                click.echo("done.", err=True)




import json

import click
import pandas as pd

from geoseeq.knex import GeoseeqNotFoundError
from multiprocessing import Pool

from .. import Organization
from .constants import *
from .fastq_utils import group_paired_end_paths, upload_fastq_pair, upload_single_fastq
from .utils import use_common_state


@click.group('upload')
def cli_upload():
    """Upload files to GeoSeeq."""
    pass

dryrun_option = click.option('--dryrun/--wetrun', default=False, help='Print what will be created without actually creating it')
overwrite_option = click.option('--overwrite/--no-overwrite', default=False, help='Overwrite existing samples and data')
module_option = lambda x: click.option('-m', '--module-name', type=click.Choice(x), default=x[0], help='Name for the module that will store the data')
private_option = click.option('--private/--public', default=True, help='Make the reads private.')
link_option = click.option(
    '--link-type',
    default='upload',
    type=click.Choice(['upload', 's3', 'ftp', 'azure', 'sra']),
    help='Link the files from a cloud storage service instead of copying them'
)
org_arg = click.argument('org_name')
project_arg = click.argument('project_name')
sample_arg = click.argument('sample_name')
module_arg = click.argument('module_name')
field_name = click.argument('field_name')


def _upload_one_sample(args):
    group, module_name, link_type, lib, filepaths, seq_length = args
    sample = lib.sample(group['sample_name']).idem()
    module = sample.analysis_result(module_name).idem()
    for field_name, path in group['fields'].items():
        field = module.field(f'{seq_length}::{field_name}').idem()
        path = filepaths[path]
        if link_type == 'upload':
            field.upload_file(path)
        else:
            field.link_file(link_type, path)
    return sample


@cli_upload.command('reads')
@use_common_state
@click.option('--cores', default=1, help='Number of uploads to run in parallel')
@click.option('--yes/--confirm', default=False, help='Skip confirmation prompts')
@click.option('--regex', default=None, help='An optional regex to use to extract sample names from the file names')
@private_option
@link_option
@module_option(['short_read::paired_end', 'short_read::single_end', 'long_read::nanopore'])
@org_arg
@project_arg
@click.argument('file_list', type=click.File('r'))
def cli_upload_reads_wizard(state, cores, yes, regex, private, link_type, module_name, org_name, project_name, file_list):
    """Upload read files to GeoSeeq.

    This command automatically groups files by their sample name, lane number and read number.

    It asks for user confirmation before creating any samples or data.

    `file_list` is a file with a list of fastq filepaths, one per line
    """
    # Set up the organization and library
    knex = state.get_knex()
    org = Organization(knex, org_name).get()
    try:
        org = Organization(knex, org_name).get()
    except GeoseeqNotFoundError:
        if not yes:
            click.confirm(f'Organization "{org_name}" does not exist. Create it?', abort=True)
        org = Organization(knex, org_name).create()
    try:
        lib = org.sample_group(project_name).get()
    except GeoseeqNotFoundError:
        if not yes:
            click.confirm(f'Project "{project_name}" does not exist. Create it?', abort=True)
        lib = org.sample_group(project_name, is_public=not private).create()

    # Find a regex that will group the files into samples and tell the user if files did not match
    
    filepaths = {line.strip().split('/')[-1]: line.strip() for line in file_list if line.strip()}
    seq_length, seq_type = module_name.split('::')[:2]
    args = {
        'filenames': list(filepaths.keys()),
        'sequence_type': seq_type,
        'sample_group_id': lib.uuid,
    }
    if regex:
        args['custom_regex'] = regex
    result = knex.post('bulk_upload/validate_filenames', json=args)
    regex = result['regex_used']
    click.echo(f'Using regex: "{regex}"', err=True)
    if result['unmatched']:
        click.echo(f'{len(result["unmatched"])} files could not be grouped.', err=True)
    else:
        click.echo('All files successfully grouped.', err=True)

    # Group files into samples, show the user what will be uploaded, and confirm
    groups = knex.post('bulk_upload/group_files', json={
        'filenames': list(filepaths.keys()),
        'sequence_type': seq_type,
        'regex': regex
    })
    for group in groups:
        click.echo(f'sample_name: {group["sample_name"]}', err=True)
        click.echo(f'  module_name: {module_name}', err=True)
        for field_name, filename in group['fields'].items():
            path = filepaths[filename]
            click.echo(f'    {seq_length}::{field_name}: {path}', err=True)
    if not yes:
        click.confirm('Do you want to upload these files?', abort=True)

    # Upload the files
    args = [(group, module_name, link_type, lib, filepaths, seq_length) for group in groups]
    with Pool(cores) as p:
        for sample in p.imap_unordered(_upload_one_sample, args):
            click.echo(f'Uploaded Sample: {sample.name}', err=True)


@cli_upload.command('file')
@use_common_state
@click.option('--yes/--confirm', default=False, help='Skip confirmation prompts')
@private_option
@link_option
@org_arg
@project_arg
@sample_arg
@module_arg
@field_name
@click.argument('file_path', type=click.Path(exists=True))
def cli_upload_file(state, yes, private, link_type, org_name, project_name, sample_name, module_name, field_name, file_path):
    """Upload a single file to a sample field."""
    knex = state.get_knex()
    try:
        org = Organization(knex, org_name).get()
    except GeoseeqNotFoundError:
        if not yes:
            click.confirm(f'Organization "{org_name}" does not exist. Create it?', abort=True)
        org = Organization(knex, org_name).create()
    try:
        lib = org.sample_group(project_name).get()
    except GeoseeqNotFoundError:
        if not yes:
            click.confirm(f'Project "{project_name}" does not exist. Create it?', abort=True)
        lib = org.sample_group(project_name, is_public=not private).create()
    sample = lib.sample(sample_name).idem()
    module = sample.analysis_result(module_name).idem()
    field = module.field(field_name).idem()
    if link_type == 'upload':
        field.upload_file(path)
    else:
        field.link_file(link_type, path)


@cli_upload.command('project-file')
@use_common_state
@click.option('--yes/--confirm', default=False, help='Skip confirmation prompts')
@private_option
@link_option
@org_arg
@project_arg
@module_arg
@field_name
@click.argument('file_path', type=click.Path(exists=True))
def cli_upload_file(state, yes, private, link_type, org_name, project_name, module_name, field_name, file_path):
    """Upload a single file to a project field."""
    knex = state.get_knex()
    try:
        org = Organization(knex, org_name).get()
    except GeoseeqNotFoundError:
        if not yes:
            click.confirm(f'Organization "{org_name}" does not exist. Create it?', abort=True)
        org = Organization(knex, org_name).create()
    try:
        lib = org.sample_group(project_name).get()
    except GeoseeqNotFoundError:
        if not yes:
            click.confirm(f'Project "{project_name}" does not exist. Create it?', abort=True)
        lib = org.sample_group(project_name, is_public=not private).create()
    module = lib.analysis_result(module_name).idem()
    field = module.field(field_name).idem()
    if link_type == 'upload':
        field.upload_file(path)
    else:
        field.link_file(link_type, path)


@cli_upload.command('metadata')
@use_common_state
@overwrite_option
@click.option('--create/--no-create', default=False)
@click.option('--update/--no-update', default=False)
@click.option('--index-col', default=0)
@click.option('--encoding', default='utf_8')
@org_arg
@project_arg
@click.argument('table', type=click.File('rb'))
def cli_metadata(state, overwrite,
                 create, update, index_col, encoding,
                 org_name, project_name, table):
    knex = state.get_knex()
    tbl = pd.read_csv(table, index_col=index_col, encoding=encoding)
    tbl.index = tbl.index.to_series().map(str)
    org = Organization(knex, org_name).get()
    lib = org.sample_group(project_name).get()
    for sample_name, row in tbl.iterrows():
        sample = lib.sample(sample_name)
        if create:
            sample = sample.idem()
        else:
            try:
                sample = sample.get()
            except Exception as e:
                click.echo(f'Sample "{sample.name}" not found.', err=True)
                continue
        new_meta = json.loads(json.dumps(row.dropna().to_dict())) 
        if overwrite or (not sample.metadata):
            sample.metadata = new_meta
            sample.idem()
        elif update:
            old_meta = sample.metadata
            old_meta.update(new_meta)
            sample.metadata = old_meta
            sample.idem()
        click.echo(sample)
